"""Defines base class from which other `Shell` objects will inherit.
"""

from abc import ABCMeta
from abc import abstractmethod

import numpy as np
from scipy import spatial

import meshio

from arnica.utils import nparray2xmf
import yamio
from yamio.ensight.gold import GeoWriter
from yamio.mesh_utils import get_local_points_and_cells

from kokiy._mesh_utils import get_conns


MAX_SPLINE_ORDER = 3
SPLINE_SMOOTHNESS = 0

AXIS_NAMES = ["time", "v", "u"]


class Shell(metaclass=ABCMeta):
    """Abstract shell.

    Args:
        shape (np.array): Shape of the shell.

    Attributes:
        shape (array-like): Shape of the shell.
        fields (dict): Nodal data.
    """

    def __init__(self, shape):
        self.shape = shape
        self.fields = {}

    def __repr__(self):
        rep = f"Shell\n-----\n\nshape : {self.shape}\ntype : {self.geom_type}"

        return rep

    def set_mask_on_shell(self, point_cloud, tol):
        """Create a mask on the shell from a point cloud.

        Args:
            point_cloud (np.array): Coordinates with shape (n, 3).
            tol (float): Tolerance.

        Notes:
            The mask value is 0 for shell points located near cloud points.
            Otherwise the mask value is 1.
        """

        # Create a KDTree from cloud points
        kdtree = spatial.KDTree(point_cloud)

        # Compute distances between cloud points and
        # the closest neighbor from the 2D shell points
        dists, _ = kdtree.query(self.xyz, k=1)

        # Mask takes 1 if the distance is above tolerance.
        # TODO: make int (breaks oms tests)
        mask = np.where(dists.reshape(self.shape) > tol, 1.0, 0.0)

        return mask

    def _dump(self, name, var_names, fields=None, exclude_internal_fields=()):
        """Dump shell geometrical properties into hdf file.

        Args:
            name (str): Name of the file to dump. Default 'shell'.
            fields (dict): Additional fields to dump. (key, `np.array`).
        """
        # TODO: rename to just `dump`?
        fields = fields or {}

        # grid
        xdmf = nparray2xmf.NpArray2Xmf(f"{name}.h5")
        xdmf.create_grid(
            np.take(self.xyz, 0, axis=-1),
            np.take(self.xyz, 1, axis=-1),
            np.take(self.xyz, 2, axis=-1),
        )

        # internal vars
        for var_name in var_names:
            xdmf.add_field(getattr(self, var_name), var_name)

        # internal fields
        for field_name, field_value in self.fields.items():
            if field_name in exclude_internal_fields:
                continue
            xdmf.add_field(field_value, field_name)

        # additional fields
        for key, value in fields.items():
            xdmf.add_field(value, key)

        xdmf.dump()

    def export_geo(self, filename, show_all=False, **kwargs):
        """Exports simplified boundary representation for visualization.

        Uses ensight gold format.

        Args:
            show_all (bool): Show all boundary edges?
            kwargs: will be passed to writer.write method.

        Notes:
            If goal is to proper export original mesh use `export_mesh`.
        """

        points = self._get_points()
        bnd_node_sets = self._get_bound_nodes_display(show_all=show_all)

        parts = {}
        for name, conns in bnd_node_sets.items():
            cells = self._reshape_simplified_bnd_conns(conns, show_all=show_all)
            new_points, new_cells = get_local_points_and_cells(points, cells)
            parts[name] = meshio.Mesh(new_points, new_cells)

        geo_writer = GeoWriter()
        geo_writer.write(filename, parts, **kwargs)

    def export_mesh(self, filename, elem_type, **kwargs):
        """Exports mesh.

        Args:
            filename (str): file name.
                Extension guides file type (any acceptable by `yamio`).
            elem_type (str): 'tri', 'quad', 'tetra', 'hexahedron'.
            kwargs: will be passed to writer.write method.

        Notes:
            `hip` does not allow 3d surface meshes.
        """
        mesh = self.get_mesh(elem_type)
        yamio.write(filename, mesh, **kwargs)

    def get_mesh(self, elem_type):
        """Returns a yamio.Mesh."""
        points, cells = self._get_points_and_cells(elem_type)
        bnd_node_groups = self._get_bound_nodes()

        return yamio.Mesh(points, cells, bnd_patches=bnd_node_groups)

    def _get_cells(self, elem_type, **kwargs):
        conns = get_conns(*self.shape, elem_type=elem_type, **kwargs)
        return [meshio.CellBlock(elem_type, conns)]

    def _get_points_and_cells(self, elem_type):
        return self._get_points(), self._get_cells(elem_type)

    @abstractmethod
    def _get_bound_nodes(self):
        pass

    @abstractmethod
    def _get_bound_nodes_display(self, **kwargs):
        pass

    @abstractmethod
    def _get_points(self):
        pass

    @abstractmethod
    def _reshape_simplified_bnd_conns(self, conns, show_all):
        pass

    @abstractmethod
    def replicate(self, shape):
        pass

    @staticmethod
    def _reshape_simplified_bnd_conns_helper(
        conns, ref_size, layered, show_all, rm_bnd_lines=False
    ):
        """Connects surface nodes by lines.

        Args:
            ref_size (int): Dimension of size of reference (i.e. number of
                nodes by layer).
            layered (bool): If False, inner lines in reference direction are
                ignored. If show_all is True, then it is ignored.
            show_all (bool): If True, all the lines are drawn.
            rm_bnd_lines (bool): If True, removes boundary lines parallel to main
                direction.
        """
        # TODO: make this a standalone function? (may be useful in different contexts)

        longi = conns.copy().reshape(-1, ref_size)
        if not layered and not show_all:
            longi = longi[[0, -1], :]
        repeats_layer = [1] + [2 for _ in range(longi.shape[1] - 2)] + [1]
        repeats = repeats_layer * longi.shape[0]
        new_conns_longi = np.repeat(longi.ravel(), repeats, axis=0).reshape(-1, 2)

        if show_all:
            azi = conns.copy().reshape(-1, ref_size).T
        else:
            azi = conns.copy().reshape(-1, ref_size)[:, [0, -1]].T
        repeats_layer = [1] + [2 for _ in range(azi.shape[1] - 2)] + [1]
        repeats = repeats_layer * azi.shape[0]
        new_conns_azi = np.repeat(azi.ravel(), repeats, axis=0).reshape(-1, 2)

        if rm_bnd_lines:
            new_conns_longi = new_conns_longi[azi.shape[0] - 1 : -(azi.shape[0] - 1), :]
            new_conns_azi = new_conns_azi[longi.shape[0] - 1 : -(longi.shape[0] - 1), :]

        return np.r_[new_conns_longi, new_conns_azi]



__version__ = "1.0.0"

# from kokiy.axishell import AxiShell
# from kokiy.cartshell import CartShell
# from kokiy.thickshell import ThickShell
# from kokiy.frustum_shell import FrustumShell
# from kokiy.torch_shell import TorchShell



import os

import numpy as np
import pytest
import h5py


def test_shell_attrs(datadir, thickshell):
    shell_type, shell = thickshell

    assert shell.geom_type == "thick"

    # shape
    np.testing.assert_allclose(shell.shape, (8, 10, 4))

    filename = os.path.join(datadir, f"{shell_type}.h5")
    with h5py.File(filename) as file:
        # mesh
        for i, var in enumerate(["x", "y", "z"]):
            print("var", var)
            np.testing.assert_allclose(
                np.take(shell.xyz, i, -1).ravel(), file[f"mesh/{var}"][()], atol=1.0e-15
            )

        # vars
        np.testing.assert_allclose(shell.dz, file["variables/dz"], atol=1.0e-15)


def test_dump(tmpdir, thickshell):
    _, shell = thickshell

    filename = os.path.join(tmpdir, "shell")

    # create a random (external) field
    data = np.random.random(shell.shape)
    data_dict = {"random_external": data}

    # create a random (internal) field
    data = np.random.random(shell.shape)
    shell.fields["random_internal"] = data

    # TODO: delete after deprecation
    shell.dump_shell(f"{filename}_bc", fields=data_dict)
    shell.dump(filename, fields=data_dict)
    size_target = np.prod(shell.shape)
    with h5py.File(f"{filename}.h5", "r") as fin:

        # mesh
        for var in ["x", "y", "z"]:
            assert fin[f"mesh/{var}"][()].size == size_target

        # other variables
        for var in ["dz", "random_external", "random_internal"]:
            assert fin[f"variables/{var}"][()].size == size_target



"""Defines a cartesian shell object.
"""

import numpy as np

from kokiy.shell_2d import Shell2D, compute_normals_from_xyz


class CartShell(Shell2D):
    r"""Base class for cartesian computational shells.

    Some attributes are based on **u** and **v**, defined as the curvilinear
    longitudinal (u) abscissa and curvilinear transversal (v) abscissa
    respectively.

    The shell is set using three corners (assuming they are orthogonal).


    Args:
        n_trans (int): Number of azimuthal shell points.
        n_longi (int): Number of longitudinal shell points.
        zero (np.array): Lower left corner of shell of shape (3,).
        umax (np.array): Corner of shell for umax of shape (3,).
        vmax (np.array): corner of shell for vmax of shape (3,).


    ::

          X umax
          ________________
          |              |            Cartesian system
          |              |            Example longi/u <=> x  and v <==> z
          |              |            zero. (0, 0, 0)
          |              |            umax. (1, 0, 0)
          |              |            vmax. (0, 0, 1)
          |              |            x - longi/u
          |              |            ^
          |              |            |
          |              |            |
          |              |            O-----> z - transvers/v
          |______________| X vmax     y
          X zero.
    """

    geom_type = "cart"

    def __init__(self, n_trans, n_longi, zero, umax, vmax):

        super().__init__(n_trans, n_longi)
        self.zero = zero
        self.umax = umax
        self.vmax = vmax
        self._build_shell(np.array(zero), np.array(umax), np.array(vmax))

    def _build_shell(self, zero, umax, vmax):
        """Build shell from geometric features.

        Notes:
            - Construct a spline used as base for extrusion from control\
                points: tck
            - Discretise the spline: shell_crest
            - Compute normal vectors for the 1D shell_crest
            - Compute r, n_x, n_r-components for 2D shell
            - Compute theta-components for 2D shell
            - Compute xyz, n_y, n_z-components for 2D shell
        """

        vec_u = umax - zero
        vec_v = vmax - zero
        self.x, self.y, self.z = _build_xyz(self.shape, zero, vec_u, vec_v)

        self.n_x, self.n_y, self.n_z = compute_normals_from_xyz(self.x, self.y, self.z)

        # Surface element
        d_u = np.linalg.norm(vec_u) / (self.shape[0] - 1)
        d_v = np.linalg.norm(vec_v) / (self.shape[1] - 1)
        self.du = np.full(self.shape, d_u)
        self.dv = np.full(self.shape, d_v)

        # Compute weight intervals in u and v directions array of shape (n_transvers, n_longi)
        self.dwu, self.dwv = self._compute_weight_interv_adim(self.du, self.dv)

    def replicate(self, shape):
        """Creates a similar instance, but possibly with different shape."""
        return CartShell(*shape, self.zero, self.umax, self.vmax)


def _build_xyz(shape, zero, vec_u, vec_v):

    u_range = np.linspace(0.0, 1.0, num=shape[0])
    v_range = np.linspace(0.0, 1.0, num=shape[1])

    v_coor, u_coor = np.meshgrid(v_range, u_range)
    # multiply coord matrix by each vec component
    u_coor = vec_u[:, np.newaxis, np.newaxis] * u_coor[np.newaxis, :, :]
    v_coor = vec_v[:, np.newaxis, np.newaxis] * v_coor[np.newaxis, :, :]
    xyz = np.stack([zero[i] + u_coor[i] + v_coor[i] for i in range(3)], axis=-1)
    x, y, z = [np.take(xyz, i, -1) for i in range(3)]
    return x, y, z



import numpy as np


def get_conns(*args, elem_type="quad", **kwargs):
    """Get connectivities given element type.

    Args:
        args: number of columns, rows and (if applicable) layers.
        elem_type (str): 'tri', 'quad', 'tetra', 'hexahedron'.
    """
    get_conns_elem = {
        "triangle": get_tri_conns,
        "quad": get_quad_conns,
        "tetra": get_tetra_conns_layered,
        "hexahedron": get_hexa_conns_layered,
    }

    return get_conns_elem[elem_type](*args, **kwargs)


def get_quad_conns(c, r, circular=False):
    dofs = [i for i in range(c * r)]
    if circular:
        dofs += dofs[:c]

    conns = []
    for j in range(c - 1 + circular):
        for i in range(r - 1):
            k = j * r
            conns.append(
                (dofs[i + k], dofs[1 + i + k], dofs[1 + i + k + r], dofs[i + k + r])
            )

    return np.array(conns)


def get_tri_conns(c, r, circular=False):
    dofs = [i for i in range(c * r)]
    if circular:
        dofs += dofs[:c]

    conns = []
    for j in range(c - 1 + circular):
        for i in range(r - 1):
            k = j * r
            conns.append((dofs[i + k], dofs[1 + i + k], dofs[1 + i + k + r]))
            conns.append((dofs[i + k], dofs[1 + i + k + r], dofs[i + k + r]))

    return np.array(conns)


def get_hexa_conns(c, r, reverse=False):
    """
    Args:
        reverse (bool): If False, assumes 4 first nodes are below last nodes.
            Affects sign of volumes (which must be positive).

    """
    n_elem_layer = r * c

    # roll due to connectivity order
    conns_square = np.roll(get_quad_conns(c, r), 1, axis=1)
    if not reverse:
        conns_square = conns_square[:, ::-1]
    conns = np.c_[conns_square, conns_square + (n_elem_layer)]

    return conns


def get_hexa_conns_layered(c, r, n, reverse=False):
    conns = get_hexa_conns(c, r, reverse=reverse)
    return _repeat_node_group_layer(c, r, n - 1, conns)


def get_tetra_conns(c, r, reverse=False):
    """
    Notes:
        Following Figure 9 of [1] to define elements.

    References:
        [1] https://www.researchgate.net/publication/271707352_Human_Facial_Soft_Tissue_Modeling_Using_Finite_Element_Method/figures?lo=1
    """
    conns_quad = get_quad_conns(c, r)
    opposite_face = conns_quad + (c * r)

    conns = []
    for i in range(conns_quad.shape[0]):
        if reverse:
            v0, v1, v3, v2 = conns_quad[i]
            v4, v5, v7, v6 = opposite_face[i]
        else:
            v4, v5, v7, v6 = conns_quad[i]
            v0, v1, v3, v2 = opposite_face[i]

        # just choosing the right dofs for each element (think about a cube)
        elem0 = [v4, v7, v5, v0]
        elem1 = [v0, v3, v2, v7]
        elem2 = [v4, v6, v7, v0]
        elem3 = [v1, v5, v7, v0]
        elem4 = [v0, v1, v3, v7]
        elem5 = [v2, v7, v6, v0]

        conns.extend([elem0, elem1, elem2, elem3, elem4, elem5])

    return np.array(conns)


def get_tetra_conns_layered(c, r, n, reverse=False):
    conns = get_tetra_conns(c, r, reverse=reverse)
    return _repeat_node_group_layer(c, r, n - 1, conns)


def _repeat_node_group_layer(c, r, repeats, group):
    """Returns corresponding nodes in successive layers."""
    group_layered = group.copy()
    for i in range(1, repeats):
        group_layered = np.r_[group_layered, group + i * c * r]

    return group_layered


def get_bound_nodes(*args, **kwargs):
    if len(args) == 2:
        return get_bound_nodes_2d(*args, **kwargs)
    elif len(args) == 3:
        return get_bounds_nodes_3d(*args, **kwargs)


def get_bound_nodes_2d(c, r, labels=("Bottom", "Right", "Top", "Left")):
    n = r * c
    bottom = np.array([i for i in range(r)])
    right = np.array([j * r + r - 1 for j in range(c)])
    top = np.array([i for i in range(n - r, n)])
    left = np.array([j * r for j in range(c)])

    node_sets = [bottom, right, top, left]
    return {label: node_set for label, node_set in zip(labels, node_sets)}


def get_bounds_nodes_3d(
    c, r, n, labels=("Bottom", "Top", "Left", "Right", "Back", "Front"), reverse=False
):
    n_bottom = c * r
    bottom = np.array([i for i in range(n_bottom)])
    top = bottom + (n - 1) * n_bottom
    left = _repeat_node_group_layer(c, r, n, np.array([i for i in range(r)]))
    right = _repeat_node_group_layer(
        c, r, n, np.array([i + n_bottom - r for i in range(r)])
    )
    back = _repeat_node_group_layer(c, r, n, np.array([j * r for j in range(c)]))
    front = _repeat_node_group_layer(
        c, r, n, np.array([j * r + r - 1 for j in range(c)])
    )

    node_sets = [bottom, top, left, right, back, front]
    if reverse:
        node_sets[0], node_sets[1] = node_sets[1], node_sets[0]

    return {label: node_set for label, node_set in zip(labels, node_sets)}



import numpy as np

from cloud2cloud import cloud2cloud


def get_uv(
    n_axe_1, n_axe_2, matrix_form=False, lims_axe_1=(0.0, 1.0), lims_axe_2=(0.0, 1.0)
):
    u, v = [
        arr.T
        for arr in np.meshgrid(
            np.linspace(*lims_axe_1, n_axe_1), np.linspace(*lims_axe_2, n_axe_2)
        )
    ]

    if matrix_form:
        return np.stack((u, v), axis=-1)
    else:
        return u, v


def get_cond_bnds(uv):
    """Returns bool array with True for boundary nodes."""
    uv_resh = uv.reshape(-1, 2)
    cond = (
        (uv_resh[:, 0] == 0)
        | (uv_resh[:, 1] == 0)
        | (uv_resh[:, 0] == 1)
        | (uv_resh[:, 1] == 1)
    )

    return cond.reshape(uv.shape[:-1])


def interp_on_uv(init_uv, data, new_uv, correct_bnds=True):

    new_data = cloud2cloud(init_uv, data, new_uv, stencil=4)

    if correct_bnds:
        init_cond_bnds = get_cond_bnds(init_uv)
        new_cond_bnds = get_cond_bnds(new_uv)

        bnds_data = cloud2cloud(
            init_uv[init_cond_bnds],
            data[init_cond_bnds],
            new_uv[new_cond_bnds],
            stencil=2,
        )
        new_data[new_cond_bnds] = bnds_data

    return new_data



import os

import numpy as np
import pytest

from kokiy.shell_2d import Shell2D

ATOL = 1e-8


# define tests
def _export_geo(datadir, tmpdir, shell, show):
    name, shell = shell

    all_str = "_all" if show else ""
    base_filename = f"{name}{all_str}.geo"

    filename = os.path.join(tmpdir, base_filename)

    shell.export_geo(filename, show_all=show)

    # created file
    with open(filename, "r") as file:
        created_txt = file.read()

    # stored file
    filename = os.path.join(datadir, "geo", base_filename)
    with open(filename, "r") as file:
        stored_txt = file.read()

    assert created_txt == stored_txt


def _replicate(shell):
    shell_2d, shell = shell

    new_shape = [3, 6]
    if not isinstance(shell, Shell2D):
        new_shape.append(3)

    new_shell = shell.replicate(new_shape)

    np.testing.assert_array_equal(new_shell.shape, new_shape)


@pytest.mark.parametrize("show_all", [False, True])
def test_export_geo(datadir, tmpdir, cartshell, show_all):
    _export_geo(datadir, tmpdir, cartshell, show_all)


@pytest.mark.parametrize("show_all", [False, True])
def test_export_geo(datadir, tmpdir, axishell, show_all):
    _export_geo(datadir, tmpdir, axishell, show_all)


@pytest.mark.parametrize("show_all", [False, True])
def test_export_geo(datadir, tmpdir, param_thickaxishell, show_all):
    _export_geo(datadir, tmpdir, param_thickaxishell, show_all)


def test_replicate_shell_2d(shell_2d):
    _replicate(shell_2d)


def test_replicate_thickshell(thickshell):
    _replicate(thickshell)


def test_mask(axishell):
    # TODO: test also on a cartshell and thickshell (backward compatibility)
    _, axishell = axishell

    point_cloud = np.array([[0.1, 0.12, 0.0]])
    tolerance = 0.04
    mask = axishell.set_mask_on_shell(point_cloud, tolerance)
    target_field = np.array(
        [
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 0, 0, 1, 1, 1, 1],
            [1, 1, 1, 0, 0, 0, 0, 1, 1, 1],
            [1, 1, 1, 0, 0, 0, 0, 1, 1, 1],
            [1, 1, 1, 0, 0, 0, 0, 1, 1, 1],
            [1, 1, 1, 0, 0, 0, 0, 1, 1, 1],
            [1, 1, 1, 1, 0, 0, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        ]
    )
    np.testing.assert_array_equal(mask, target_field)



import numpy as np
from pytest import approx
from kokiy.axishell import AxiShell


def test_angle_min():
    ctrl_pts_x = (0.0, 0.2)
    ctrl_pts_r = (0.12, 0.12)
    n_azi, n_long = 8, 10
    angle = 40.0
    angle_min = 0.0
    axishell = AxiShell(
        n_azi, n_long, angle, ctrl_pts_x, ctrl_pts_r, angle_min=angle_min
    )

    print(
        "1 \n",
        axishell.theta,
        "2 \n",
        np.min(axishell.theta),
        "3 \n",
        np.rad2deg(np.min(axishell.theta)),
    )

    assert np.rad2deg(np.min(axishell.theta)) == approx(0.0)
    assert np.rad2deg(np.max(axishell.theta)) == approx(40.0)



"""Defines cut plane tools.
"""

import numpy as np

from arnica.utils.vector_actions import yz_to_theta
from cloud2cloud import CloudInterpolator

from kokiy.axishell import AxiShell
from kokiy.cartshell import CartShell

# TODO: need to review how general this functions are


__all__ = [
    "infer_shell_from_xyz",
    "axishell_create_structured_base",
    "cartshell_create_structured_base",
    "interpolate_solutions_on_shell",
    "shell_geom_type",
    "shell_repr",
]


def infer_shell_from_xyz(mesh_xyz, geom_type, shape, bnd_uv=None, bnd_angles=None):
    """Build a shell from an AVBP Cut.

    Args:
        geom_type (str): Geometry type. Either "cart" or "axicyl".
        shape (array-like): Discretization of the shell. Shape (nu, nv).
        bnd_uv (array-like): Limits shell umin, umax, vmin, vmax. Shape (4,).
    """
    if bnd_angles is not None:
        raise DeprecationWarning("BNd_angle is not used anymore, Use bnd_uv")

    # Build shell
    if geom_type == "axicyl":
        shell = axishell_create_structured_base(shape[0], shape[1], mesh_xyz, bnd_uv)
    elif geom_type == "cart":
        shell = cartshell_create_structured_base(shape[0], shape[1], mesh_xyz)
    return shell


def axishell_create_structured_base(n_longi, n_azi, mesh, bnd_uv=None):
    """Obtain the base in a structured mesh.

    Assumes the mesh is the x_axis rotated extrusion of an (x,r) curve.

    Args:
        n_longi (int): Number of radial points of the shell.
        n_azi (int): Number of azimuthal points of the shell.
        mesh (array-like): Coordinates of the cut mesh of shape (n, 3).
        bnd_uv (array-like): Limits shell umin, umax, vmin, vmax. Shape (4,).

    Returns:
        AxiShell: An axysymmetric computational shell.
    """

    x_pos = mesh[:, 0].mean()
    r_vals = np.hypot(mesh[:, 1], mesh[:, 2])
    theta = np.rad2deg(yz_to_theta(mesh[:, :]))

    if bnd_uv is None:
        bnd_uv = [None, None, None, None]
    if bnd_uv[0] is None:
        bnd_uv[0] = r_vals.min()
    if bnd_uv[1] is None:
        bnd_uv[1] = r_vals.max()
    if bnd_uv[2] is None:
        bnd_uv[2] = theta.min()
    if bnd_uv[3] is None:
        bnd_uv[3] = theta.max()

    x_ctrl_pts = [x_pos, x_pos]
    r_ctrl_pts = [bnd_uv[0], bnd_uv[1]]
    tmin = bnd_uv[2]
    tmax = bnd_uv[3]

    shell = AxiShell(
        n_azi,
        n_longi,
        tmax - tmin,  # angle_range,
        x_ctrl_pts,  # x_vals[corners_idx],
        r_ctrl_pts,  # r_vals[corners_idx],
        tmin,  # angle_min,
    )

    return shell


def cartshell_create_structured_base(n_longi, n_trans, mesh):
    """Obtain the base in a structured mesh.

    Assumes the path is square, aligned with y and z, with x = constant.

    Args:
        n_longi (int): Number of longitudinal points of the shell.
        n_trans (int): Number of transversal points of the shell.
        mesh (array-like): Coordinates of the cut mesh of shape (n, 3).

    Returns:
        CartShell: A cartesian computational shell.
    """

    x_vals, y_vals, z_vals = mesh.T

    zero = (x_vals.mean(), y_vals.min(), z_vals.min())
    v_max = (x_vals.mean(), y_vals.max(), z_vals.min())
    u_max = (x_vals.mean(), y_vals.min(), z_vals.max())
    shell = CartShell(
        n_trans,
        n_longi,
        zero,
        u_max,
        v_max,
    )

    return shell


def interpolate_solutions_on_shell(
    shell, mesh, raw_data, stencil=4, function=None, limitsource=None
):
    """Interpolates solutions on a shell.

    Args:
        shell (Shell): A `Shell` object.
        mesh (np.array): Mesh with shape (n_pts, 3).
        data_dict (dict or np.array): Arrays shape = (n_steps, n_pts) or (n_pts,).
            Must contain keys if `np.array` (prefer `dict` for this case).

    Notes:
        Check cloud2cloud.CloudInterpolator for meaning of other arguments.
    """
    if function is None:
        function = lambda x: np.power(x, 4)

    # manipulate input
    is_dict = type(raw_data) == dict
    data_dict = raw_data if is_dict else _from_np_struct_to_dict(raw_data)

    # create interpolator
    interpolator = CloudInterpolator(
        mesh.reshape(-1, 3).T,
        shell.xyz.reshape(-1, 3).T,
        stencil=stencil,
        function=function,
        limitsource=limitsource,
    )

    n = mesh.shape[0]
    results_dict = {}
    for key, array in data_dict.items():
        if array.shape[0] == n:
            results_dict[key] = interpolator.interp(array).reshape(*shell.shape)
        else:
            results_dict[key] = interpolator.interp(array.T).T.reshape(-1, *shell.shape)

    # manipulate output
    results = results_dict if is_dict else _from_dict_to_np_struct(results_dict)

    return results


def interpolate_solution_on_shell(shell, mesh, array, **kwargs):
    """Interpolates a solution on a shell.

    Args:
        shell (Shell): A `Shell` object.
        mesh (np.array): Mesh with shape (n_pts, 3).
        array (np.array): Shape = (n_steps, pts) or (n_pts,).
        kwargs : cloud2cloud.CloudInterpolator parameters.
    """
    return interpolate_solutions_on_shell(shell, mesh, {"data": array}, **kwargs)[
        "data"
    ]


def shell_geom_type(shell):
    """Return the shell geom_type.

    .. deprecated:: 0.2.0
        Use `shell.geom_type` instead.
    """
    return shell.geom_type


def shell_repr(shell):
    """Return a shell description.

    .. deprecated:: 0.2.0
        Use `repr(shell)` instead.
    """
    return repr(shell)


def _from_np_struct_to_dict(array):
    return {key: array[key] for key in array.dtype.names}


def _from_dict_to_np_struct(my_dict, type_=np.float32):
    keys = list(my_dict.keys())
    dtype = dict(names=keys, formats=[type_] * len(keys))
    shape = my_dict[keys[0]].shape

    array = np.empty(shape, dtype=dtype)
    for key, value in my_dict.items():
        array[key] = value

    return array



import os

import numpy as np
import h5py
import pytest

from kokiy.cartshell import CartShell


ATOL = 1e-8


@pytest.mark.parametrize(
    "load_shell_case",
    [
        ("cartshell", "cart"),
        ("axishell", "axicyl"),
        ("torch_shell", "torch"),
    ],
    indirect=True,
)
def test_shell_attrs(datadir, load_shell_case):
    # frustum shell is tested differently
    (shell_type, shell), geom_type = load_shell_case

    assert shell.geom_type == geom_type

    # shape
    np.testing.assert_allclose(shell.shape, (8, 10))

    filename = os.path.join(datadir, f"{shell_type}.h5")

    target_vars = [
        "theta",
        "n_x",
        "n_y",
        "n_z",
        "du",
        "u",
        "v",
        "dv",
        "dwu",
        "dwv",
        "surf",
    ]
    if shell_type not in ["frustumshell", "torchshell"]:
        target_vars.extend(["n_r"])

    with h5py.File(filename) as file:

        # mesh
        for i, var in enumerate(["x", "y", "z"]):
            np.testing.assert_allclose(
                np.take(shell.xyz, i, -1).ravel(), file[f"mesh/{var}"][()], atol=1.0e-15
            )

        # radius (different name)
        np.testing.assert_allclose(shell.rad, file["variables/r"][()], atol=1.0e-15)

        # other vars
        for var in target_vars:
            print("var", var)
            np.testing.assert_allclose(
                getattr(shell, var), file[f"variables/{var}"][()], atol=1.0e-15
            )


def test_dump(tmpdir, shell_2d):
    shell_type, shell = shell_2d

    filename = os.path.join(tmpdir, "shell")

    # create a random (external) field
    data = np.random.random(shell.shape)
    data_dict = {"random_external": data}

    # create a random (internal) field
    data = np.random.random(shell.shape)
    shell.fields["random_internal"] = data

    # dump shell
    # TODO: delete after deprecation
    shell.dump_shell(f"{filename}_bc", fields=data_dict)
    shell.dump(filename, fields=data_dict)
    size_target = shell.shape[0] * shell.shape[1]

    target_vars = [
        "r",
        "theta",
        "n_x",
        "n_y",
        "n_z",
        "du",
        "u",
        "v",
        "dv",
        "dwu",
        "dwv",
        "surf",
        "random_external",
        "random_internal",
    ]
    if shell_type not in ["frustumshell", "torchshell"]:
        target_vars.extend(["n_r"])

    with h5py.File(f"{filename}.h5", "r") as fin:
        # mesh
        for var in ["x", "y", "z"]:
            assert fin[f"mesh/{var}"][()].size == size_target

        # other variables
        for var in target_vars:
            assert fin[f"variables/{var}"][()].size == size_target


def test_average_over_dir(axishell):
    # TODO: is this test meaningful enough?
    _, axishell = axishell

    np.random.seed(1)
    target_field = np.random.random((3, axishell.shape[0], axishell.shape[1]))

    # with scale
    average = axishell.average_on_shell_over_dirs(target_field, ["v", "time"])
    target_val = np.array(
        [
            4.66739841e-05,
            1.27688882e-04,
            1.14211735e-04,
            1.08330533e-04,
            9.72293499e-05,
            1.22829723e-04,
            1.06270736e-04,
            1.32217197e-04,
            1.19892968e-04,
            6.73650501e-05,
        ]
    )
    np.testing.assert_allclose(average, target_val, atol=ATOL)

    # without scale
    average = axishell.average_on_shell_over_dirs(
        target_field, ["v", "time"], scale=False
    )
    target_val = np.array(
        [
            0.41297985,
            0.55279941,
            0.48937754,
            0.4651728,
            0.40128855,
            0.51424186,
            0.43755768,
            0.56629453,
            0.53040696,
            0.59220561,
        ]
    )
    np.testing.assert_allclose(average, target_val, atol=ATOL)

    # in case of bad definition of directions
    with pytest.raises(ValueError):
        axishell.average_on_shell_over_dirs(target_field, ["v", "k"])


def test_integrate_over_dir():
    # create a very simple shell
    n_trans, n_longi = 5, 6
    zero = np.zeros(3)
    umax = np.array([1.0, 0.0, 0.0])
    vmax = np.array([0.0, 1.0, 0.0])
    shell = CartShell(n_trans, n_longi, zero, umax, vmax)

    # create a field
    def field(x, t):
        return x[:, 0] * t  # just a linear field

    ts = [1.0, 2.0, 3.0]
    n_t = len(ts)
    fields = np.array(
        [field(shell.xyz.reshape(-1, 3), t).reshape(shell.shape) for t in ts]
    )

    # check average u
    average_u = shell.operate_on_shell_over_dirs(
        fields, directions=["u"], operator=np.mean
    )
    assert average_u.shape == (n_t, n_trans)
    for k in range(n_t):
        np.testing.assert_allclose(average_u[k], fields[k, :, 0])

    # check average v
    average_v = shell.operate_on_shell_over_dirs(
        fields, directions=["v"], operator=np.mean
    )
    assert average_v.shape == (n_t, n_longi)
    for k, t in enumerate(ts):
        assert (average_v[k][0] - t / 2) < ATOL

    # check average uv
    average_uv = shell.operate_on_shell_over_dirs(
        fields, directions=["v", "u"], operator=np.mean
    )
    assert average_uv.shape == (n_t,)
    for k, t in enumerate(ts):
        assert (average_uv[k] - t / 2) < ATOL

    # check average t (order do not matter)
    average_t = shell.operate_on_shell_over_dirs(
        fields, directions=["v", "u", "time"], operator=np.mean
    )
    assert (average_t - 1.0) < ATOL

    # check var u
    var_u = shell.operate_on_shell_over_dirs(fields, directions=["u"], operator=np.var)
    assert np.sum(np.abs(var_u)) == 0.0

    # check var v
    var_v = shell.operate_on_shell_over_dirs(fields, directions=["v"], operator=np.var)
    for k, var in enumerate([0.125, 0.5, 1.125]):
        assert (var_v[k][0] - var) < ATOL

    # check var v time
    var_vt = shell.operate_on_shell_over_dirs(
        fields, directions=["v", "time"], operator=np.var
    )
    assert var_vt.shape == (n_longi,)
    assert (np.mean(var_vt) - 0.17013889) < ATOL
    assert (var_vt[0] - 0.17013889) < ATOL

    # check var time v
    var_tv = shell.operate_on_shell_over_dirs(
        fields, directions=["time", "v"], operator=np.var
    )
    assert var_tv.shape == (n_longi,)
    assert (np.mean(var_tv) - 0.06041667) < ATOL
    assert (var_tv[0] - 0.06041667) < ATOL


def test_plot(tmpdir, shell_2d):
    _, shell = shell_2d
    shell.plot(os.path.join(tmpdir, "fig.png"))


def test_invert_normals(shell_2d):
    shell_type, shell = shell_2d

    normal_vars = ["n_x", "n_y", "n_z"]
    if shell_type not in ["frustumshell", "torchshell"]:
        normal_vars.append("n_z")

    target_val = -np.array([getattr(shell, var) for var in normal_vars])

    shell.invert_normals()
    new_normals = np.array([getattr(shell, var) for var in normal_vars])

    np.testing.assert_allclose(new_normals, target_val)


def test_baking(tmpdir, axishell):
    _, shell_2d = axishell

    # test add_curviwidth
    width_profile = (
        (0.0, -0.04),
        (0.2, -0.04),
        (0.4, -0.07),
        (0.6, -0.04),
        (0.8, -0.04),
    )
    shell_2d.add_curviwidth("width_profile", width_profile)
    target_val = np.array(
        [
            -0.04,
            -0.04,
            -0.04333333,
            -0.06,
            -0.06333333,
            -0.04666667,
            -0.04,
            -0.04,
            -0.04,
            -0.04,
        ]
    )
    np.testing.assert_allclose(
        shell_2d.width_matrix["width_profile"][0], target_val, atol=ATOL
    )

    # test addition of same name (stateful)
    shell_2d.add_curviwidth("width_profile", width_profile)
    np.testing.assert_allclose(
        shell_2d.width_matrix["width_profile"][0], 2 * target_val, atol=ATOL
    )

    # test bake_millefeuille
    cake = shell_2d.bake_millefeuille("width_profile", 4)

    # create a random (external) field
    data_dict = {"random_external": np.random.random(cake.shape)}

    # check dict like behavior
    cake["random_internal"] = np.random.random(cake.shape)

    filename = os.path.join(tmpdir, "shell")
    cake.dump(filename=filename, fields=data_dict)

    size_target = np.prod(cake.shape)
    with h5py.File(f"{filename}.h5", "r") as fin:

        # assert xyz do not exist
        with pytest.raises(KeyError):
            fin["variables/xyz"]

        # mesh
        for var in ["x", "y", "z"]:
            assert fin[f"mesh/{var}"][()].size == size_target

        # other variables
        for var in ["dz", "random_external", "random_internal"]:
            assert fin[f"variables/{var}"][()].size == size_target

    # test replicate
    new_shape = (4, 6, 3)
    new_cake = cake.replicate(shape=new_shape)
    np.testing.assert_array_equal(new_cake.shape, new_shape)


def test_rad_theta():
    # create a very simple shell
    n_trans, n_longi = 3, 3
    zero = np.array([1.0, 1.0, 0.0])
    umax = np.array([2.0, 1.0, 0.0])
    vmax = np.array([1.0, 2.0, 0.0])
    shell_xy = CartShell(n_trans, n_longi, zero, umax, vmax)

    vel_y = np.full(shell_xy.shape, 1.0)
    vel_z = np.full(shell_xy.shape, 0.0)
    vel_rad, vel_tht = shell_xy.rad_theta_components(vel_y, vel_z)
    assert np.allclose(vel_rad, 1.0)
    assert np.allclose(vel_tht, 0.0)

    vel_y = np.full(shell_xy.shape, 0.0)
    vel_z = np.full(shell_xy.shape, 1.0)
    vel_rad, vel_tht = shell_xy.rad_theta_components(vel_y, vel_z)
    assert np.allclose(vel_rad, 0.0)
    assert np.allclose(vel_tht, 1.0)



import pytest
import numpy as np

from kokiy.torch_shell import TorchShell


def test_ecc_validation():
    n_azi, n_longi = 8, 10
    pt_left = np.array((0.0, 0.0, 0.0))
    pt_right = np.array((0.0, 0.0, 1.0))
    radius_left = 0.3
    radius_right = 1.2

    # null exception
    ecc_pt = np.array((0.0, 0.0, 1.0))
    with pytest.raises(Exception):
        TorchShell(n_azi, n_longi, pt_left, pt_right, radius_left, radius_right, ecc_pt)

    # not perpendicular exception
    ecc_pt = np.array((1.0, 0.0, 0.0))
    with pytest.raises(Exception):
        TorchShell(n_azi, n_longi, pt_left, pt_right, radius_left, radius_right, ecc_pt)



import os
import json

import pytest
import numpy as np

from kokiy._mesh_utils import get_conns
from kokiy._mesh_utils import get_bound_nodes


def read_conns_data(datadir):

    filename = os.path.join(datadir, "conns.json")
    with open(filename, "r") as file:
        data = json.load(file)

    return data


@pytest.mark.parametrize("reverse", [False, True])
@pytest.mark.parametrize("elem_type", ["quad", "triangle"])
def test_conns_2d(datadir, elem_type, reverse):
    data = read_conns_data(datadir)

    c, r = 3, 4

    elem_type = "quad"
    conns = get_conns(c, r, elem_type=elem_type)

    expected_conns = data["conns_2d"][elem_type]

    np.testing.assert_array_equal(conns, expected_conns)


@pytest.mark.parametrize("reverse", [True, False])
@pytest.mark.parametrize("elem_type", ["tetra", "hexahedron"])
def test_conns_3d(datadir, elem_type, reverse):
    data = read_conns_data(datadir)

    c, r, n = 3, 4, 3
    conns = get_conns(c, r, n, elem_type=elem_type, reverse=reverse)

    expected_conns = data["conns_3d"][elem_type][str(reverse).lower()]

    np.testing.assert_array_equal(conns, expected_conns)


def test_bound_nodes_2d():
    c, r = 3, 4
    bnd_nodes = get_bound_nodes(c, r)

    # assumes dict as ordered
    exp_nodes = [[0, 1, 2, 3], [3, 7, 11], [8, 9, 10, 11], [0, 4, 8]]

    for bnd_node_group, exp_node_group in zip(bnd_nodes.values(), exp_nodes):
        np.testing.assert_array_equal(bnd_node_group, exp_node_group)


def test_bound_nodes_3d():
    c, r, n = 4, 3, 3

    # normal order
    bnd_nodes = get_bound_nodes(c, r, n, reverse=False)

    # assumes dict as ordered
    exp_nodes = [
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        [24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35],
        [0, 1, 2, 12, 13, 14, 24, 25, 26],
        [9, 10, 11, 21, 22, 23, 33, 34, 35],
        [0, 3, 6, 9, 12, 15, 18, 21, 24, 27, 30, 33],
        [2, 5, 8, 11, 14, 17, 20, 23, 26, 29, 32, 35],
    ]

    for bnd_node_group, exp_node_group in zip(bnd_nodes.values(), exp_nodes):
        np.testing.assert_array_equal(bnd_node_group, exp_node_group)

    # reverse order (bottom swaps with top)
    bnd_nodes = get_bound_nodes(c, r, n, reverse=True)

    exp_nodes[0], exp_nodes[1] = exp_nodes[1], exp_nodes[0]
    for bnd_node_group, exp_node_group in zip(bnd_nodes.values(), exp_nodes):
        np.testing.assert_array_equal(bnd_node_group, exp_node_group)



"""Defines base class from which other 2D `Shell` objects will inherit.
"""

from abc import ABCMeta

import numpy as np
import matplotlib.pyplot as plt
import meshio

from kokiy.shell import Shell
from kokiy.thickshell import CakeDict
from kokiy.thickshell import create_width_profile
from kokiy._mesh_utils import get_bound_nodes
from kokiy._structured import get_uv


AXIS_NAMES = ["time", "v", "u"]


class Shell2D(Shell, metaclass=ABCMeta):
    """Abstract shell.

    Args:
        n_axe_1 (int): Number of x/longitudinal shell points.
        n_axe_2 (int): Number of y/azimuthal shell points.

    Attributes:
        shape (array-like): Shape of the shell: (`n_axe_1`, `n_axe_2`).
        rad (np.array): rad matrix of shape `self.shape`.
        theta (np.array): theta matrix of shape `self.shape`.
        x (np.array): x matrix of shape `self.shape`.
        y (np.array): y matrix of shape `self.shape`.
        z (np.array): z matrix of shape `self.shape`.
        n_x (np.array): x-normal matrix of shape `self.shape`.
        n_r (np.array): r-normal matrix of shape `self.shape`.
        n_y (np.array): y-normal matrix of shape `self.shape`.
        n_z (np.array): z-normal matrix of shape `self.shape`.
        du (np.array): u-direction spacing matrix of shape `self.shape`.
        dv (np.array): v-direction spacing matrix of shape `self.shape`.
        u (np.array): u-direction adimensional matrix of shape `self.shape`.
        v (np.array): v-direction adimensional matrix of shape `self.shape`.
        abs_curv (np.array): Central curvilinear abscissa array of shape (`self.shape[1]`,).
        dwu (np.array): Weighted u-direction spacing matrix of shape `self.shape`.
        dwv (np.array): Weighted v-direction spacing matrix of shape `self.shape`.
        surf (np.array): Weighted surface matrix of shape `self.shape`.
    """

    def __init__(self, n_axe_1, n_axe_2, lims_u=(0.0, 1.0), lims_v=(0.0, 1.0)):

        super().__init__((n_axe_1, n_axe_2))

        self.axis_names = AXIS_NAMES
        self.width_matrix = {}
        self.u, self.v = get_uv(n_axe_1, n_axe_2, lims_axe_1=lims_u, lims_axe_2=lims_v)

    @property
    def uv(self):
        """uv matrix of shape `self.shape`."""
        return np.stack((self.u, self.v), axis=-1)

    @property
    def xyz(self):
        """xyz matrix of shape (`*self.shape`, 3)."""
        return np.stack((self.x, self.y, self.z), axis=-1)

    @property
    def n_xyz(self):
        """Normals with same shape as `xyz`."""
        return np.stack((self.n_x, self.n_y, self.n_z), axis=-1)

    @property
    def surf(self):
        """Surface element [m2]"""
        return np.multiply(self.dwu, self.dwv)

    @property
    def rad(self):
        """radius of point [m]"""
        return np.hypot(self.y, self.z)

    @property
    def theta(self):
        """azimuth of point [rad]"""
        return np.arctan2(self.z, self.y)

    @property
    def n_r(self):
        """radial component of normal [-]"""
        return self.n_y * np.cos(self.theta) + self.n_z * np.sin(self.theta)

    @property
    def abs_curv(self):
        """Curvilinear abcissa on first dimension U [m]"""
        return np.cumsum(np.take(self.du, 0, 0)) - self.du[0, 0]

    @property
    def u_adim(self):
        """
        .. deprecated:: 0.2.0
            Use `u` instead.
        """
        return self.u

    @property
    def v_adim(self):
        """
        .. deprecated:: 0.2.0
            Use `v` instead.
        """
        return self.v

    def dump(self, filename="shell", fields=None):
        """Dump shell geometrical and physical properties into hdf file.

        Args:
            filename (str): Name of the file to dump.
            fields (dict): Additional fields to dump. (key, `np.array`).
        """
        fields = fields or {}

        var_names = self._get_dump_var_names()

        # due to name inconsistency
        fields.update({"r": self.rad})

        # dump
        self._dump(filename, var_names, fields=fields)

    def dump_shell(self, name="shell", fields=None):
        """Dump shell geometrical properties into hdf file.

        Args:
            name (str): Name of the file to dump. Default 'shell'.
            fields (dict): Additional fields to dump. (key, `np.array`).

        .. deprecated:: 0.2.0
            Use `dump` instead.
        """
        self.dump(filename=name, fields=fields)

    def add_curviwidth(self, label, points):
        """Add a 2D width matrix of shell shape extruded from points spline.

        Args:
            label (str): Label of the width matrix.
            points (array-like): Coordinates with shape (n, 2).

        Notes:
            This method makes this object stateful. Avoid it if you can.
        """
        width_profile = create_width_profile(points, self.shape)
        if label in self.width_matrix:
            self.width_matrix[label] += width_profile
        else:
            self.width_matrix[label] = width_profile

    def bake_millefeuille(self, width_matrix_label, n_layers, shift=0.0):
        """Create a millefeuille-like shell.

        Extrude a 2D shell in the normal direction up
        pointwise height given by "width_matrix_label" matrix.

        Args:
            width_matrix_label (str): Label of the width matrix.
            n_layers (int): Number of layer for extrusion.
            shift (float): Additional depth.

        Returns:
            ThickShell

        Notes:
            Use `Thickshell.bake_from_shell` when possible.

        "Bon appetit!"
        """
        width_matrix = self.width_matrix[width_matrix_label]
        return CakeDict.bake_from_shell(
            self, width_matrix, n_layers=n_layers, shift=shift
        )

    def average_on_shell_over_dirs(self, variable, directions, scale=True):
        """Performs an integration (average) over one or multiple directions.

        Args:
            variable (np.array): Data to be averaged of shape (n_time, n_v, n_u).
            directions (array-like): A list() of directions on which the average process
                is to be performed. Contains keywords from ['time', 'v', 'u'].
            scale (bool): Keeps nominal averaging (if `False`) or takes into account
                surface element (if `True`).

        Returns:
            np.array: Averaged data on given directions.
        """
        if scale:
            variable = np.multiply(variable, self.surf)

        return self.operate_on_shell_over_dirs(variable, directions, operator=np.mean)

    def operate_on_shell_over_dirs(self, variable, directions, operator=np.mean):
        """Applies an operator over one or multiple direction.

        Notes:
            Directions are not commutative for non-linear operators.
        """

        # check valid directions
        if not (
            len(directions) > 0 and all(dir_ in self.axis_names for dir_ in directions)
        ):
            mess = "Integrating directions do not conform to criteria\n"
            mess += "It should be a list containing one of or all items\n"
            for axis in self.axis_names:
                mess += axis + ", "
            raise ValueError(mess)

        # operate
        dirs = self.axis_names.copy()
        for dir_ in directions:
            index = dirs.index(dir_)
            variable = operator(variable, axis=index)
            dirs.pop(index)

        return variable

    def plot(self, savefile=None):
        """Plot 2D curve of the `Shell`.

        Args:
            savefile (str): filename.
                If None, the figure is only plotted.
        """

        plt.plot(self.xyz[0, :, 0], self.xyz[0, :, 1])
        plt.title("Shell")
        plt.xlabel("x")
        plt.ylabel("y")

        if savefile is None:
            plt.show()
        else:
            plt.savefig(savefile)

    def _invert_normals(self, normal_vars=("n_x", "n_y", "n_z")):
        for var in normal_vars:
            setattr(self, var, -getattr(self, var))

    def invert_normals(self):
        """Invert directions of normal vectors."""
        self._invert_normals()

    def rad_theta_components(self, vect_y, vect_z):
        """Computes the radial and azimuthal components of a vect y, z."""
        # TODO: example of use case?

        unit_rad_y = self.y / self.rad
        unit_rad_z = self.z / self.rad

        unit_tgt_y = -unit_rad_z
        unit_tgt_z = unit_rad_y

        vect_rad = vect_y * unit_rad_y + vect_z * unit_rad_z
        vect_tgt = vect_y * unit_tgt_y + vect_z * unit_tgt_z

        return vect_rad, vect_tgt

    def _get_points(self):
        return self.xyz.reshape(-1, 3)

    def _reshape_simplified_bnd_conns(self, conns, **kwargs):
        size = conns.size
        if size == self.shape[0] or size == self.shape[1]:  # not inner
            new_conns = conns.copy()
            repeats = [1] + [2 for _ in range(new_conns.shape[0] - 2)] + [1]
            new_conns = np.repeat(new_conns, repeats, axis=0).reshape(-1, 2)
        else:
            new_conns = self._reshape_simplified_bnd_conns_helper(
                conns, self.shape[1], False, True, rm_bnd_lines=True
            )

        return [meshio.CellBlock("line", new_conns)]

    def _get_bound_nodes_display(self, show_all=False):
        bnd_nodes = self._get_bound_nodes()
        if show_all:
            # all nodes, not only inner (then connections of bnd are ignored)
            bnd_nodes["Inner"] = np.array([i for i in range(np.prod(self.shape))])

        return bnd_nodes

    def _get_bound_nodes(self):
        return get_bound_nodes(*self.shape)

    # def _compute_abs_curv(self, du):
    #     return np.cumsum(np.take(du, 0, 0)) - du[0, 0]

    def _compute_weight_interv_adim(self, du, dv):
        dwu = du.copy()
        dwu[:, (0, -1)] /= 2

        dwv = dv.copy()
        dwv[(0, -1), :] /= 2

        return dwu, dwv

    # def _compute_surf(self, dwu, dwv):
    #     return np.multiply(dwu, dwv)

    def _get_dump_var_names(self):
        return [
            "theta",
            "n_x",
            "n_y",
            "n_z",
            "n_r",
            "du",
            "dv",
            "u",
            "v",
            "dwu",
            "dwv",
            "surf",
        ]


def compute_normals_from_xyz(x, y, z, perio_u=False, perio_v=False):
    """Compute normals from coordinates"""
    du_vect = np.stack(
        [
            np.gradient(x, axis=0, edge_order=2),
            np.gradient(y, axis=0, edge_order=2),
            np.gradient(z, axis=0, edge_order=2),
        ],
        axis=2,
    )
    dv_vect = np.stack(
        [
            np.gradient(x, axis=1, edge_order=2),
            np.gradient(y, axis=1, edge_order=2),
            np.gradient(z, axis=1, edge_order=2),
        ],
        axis=2,
    )

    dn = np.cross(du_vect, dv_vect, axisc=2)

    if perio_u:
        avg = 0.5 * (dn[0, :, :] + dn[-1, :, :])
        dn[0, :, :] = avg
        dn[-1, :, :] = avg

    if perio_v:
        avg = 0.5 * (dn[:, 0, :] + dn[:, -1, :])
        dn[:, 0, :] = avg
        dn[:, -1, :] = avg

    dn /= np.linalg.norm(dn, axis=2)[:, :, np.newaxis]

    # dn= clean_zeros(dn)
    return dn[:, :, 0], dn[:, :, 1], dn[:, :, 2]


def clean_zeros(arr, tol=1e-12):

    return np.where(np.abs(arr) < tol, 0, arr)



"""Defines an x-axisymmetric shell object.
"""

import numpy as np
from scipy import interpolate

from kokiy.shell_2d import Shell2D, compute_normals_from_xyz

MAX_SPLINE_ORDER = 3
SPLINE_SMOOTHNESS = 0


class AxiShell(Shell2D):
    r"""An x-axisymmetric computational shell.

    Some attributes are based on **u** and **v**, defined as the curvilinear
    longitudinal (x/r) abscissa and curvilinear azimutal (theta) abscissa
    respectively.


    Args:
        n_azi (int): Number of azimuthal shell points.
        n_longi (int): Number of longitudinal shell points.
        angle (float): Range angle of the axicylindrical geometry.
        ctrl_pts_x (np.array): x-coordinates of the points defining the spline
            of shape (n,).
        ctrl_pts_r (np.array): r-coordinates of the points defining the spline
            of shape (n,).
        angle_min (float): Minimum angle of the shell.


    ::

                      (x_n, r_n)
                _____X_____
         ___----     |     ----___     Cylindrical system
        \            |            /    Example longi/u <=> r
         \           |           /
          \          |          /      r - longi/u
           \         |         /       ^
            \        | (x_0, r_0)      |
             \      _X_      /         |
              \__---   ---__/          X-----> theta - azi/v
                                      x

              ................
          ...                 ..       Cylindrical system
        ..                     .       Example longi/u <=> x/r
        .                      .
        .      ................         r
        .     .                         ^
         .     ...................      |
         ..                             |
            ......................      X-----> x
                          <--      theta - azi/v
                     x/r - longi/u
    """

    geom_type = "axicyl"

    def __init__(self, n_azi, n_longi, angle, ctrl_pts_x, ctrl_pts_r, angle_min=None):
        super().__init__(n_azi, n_longi)

        self.ctrl_pts_x = ctrl_pts_x
        self.ctrl_pts_r = ctrl_pts_r
        self.angle = angle
        self.angle_min = angle_min

        self._build_shell()

    def _build_shell(self):
        """Builds shell from geometric features.

        Notes:
            - Construct a spline used as base for extrusion
              from control points : tck
            - Discretise the spline : shell_crest
            - Compute normal vectors for the 1D shell_crest
            - Compute r,n_x,n_r-components for 2D shell
            - Compute theta-components for 2D shell
            - Compute xyz,n_y,n_z-components for 2D shell
        """

        # Construct Shell Crest
        shell_crest = _compute_shell_crest(
            self.ctrl_pts_x, self.ctrl_pts_r, self.shape[1]
        )

        # Compute radius and theta matrices of shape (n_azi, n_longi)
        rot_angle = 0
        if self.angle_min is not None:
            rot_angle = 0.5 * self.angle + self.angle_min

        min_theta = (rot_angle - 0.5 * self.angle) * np.pi / 180
        max_theta = (rot_angle + 0.5 * self.angle) * np.pi / 180
        theta_vec = np.linspace(min_theta, max_theta, num=self.shape[0])
        rad, theta = np.meshgrid(shell_crest[1], theta_vec)

        # Compute xyz matrix of shape (n_azi, n_longi, 3)
        self.x = np.tile(shell_crest[0], (self.shape[0], 1))
        self.y = rad * np.cos(theta)
        self.z = rad * np.sin(theta)

        perio_u = False
        if self.angle == 360.0:
            perio_u = True
        self.n_x, self.n_y, self.n_z = compute_normals_from_xyz(
            self.x, self.y, self.z, perio_u=perio_u
        )

        # Compute du, dv matrix of shape (n_azi, n_longi)
        self.du = np.pad(
            np.sqrt(np.diff(self.x, axis=1) ** 2 + np.diff(rad, axis=1) ** 2),
            ((0, 0), (1, 0)),
            "edge",
        )
        self.dv = rad * np.pad(np.diff(theta, axis=0), ((1, 0), (0, 0)), "edge")

        # Compute weight intervals in u and v directions array of shape (n_transvers, n_longi)
        self.dwu, self.dwv = self._compute_weight_interv_adim(self.du, self.dv)

    def replicate(self, shape):
        """Creates a similar instance, but possibly with different shape."""
        return AxiShell(
            *shape,
            self.angle,
            self.ctrl_pts_x,
            self.ctrl_pts_r,
            angle_min=self.angle_min,
        )


def _compute_shell_crest(ctrl_pts_1, ctrl_pts_2, n):
    """Build the shell crest of shape (2, self.shape[1]) and respecting
    controls points.
    """

    # Build crest
    spline_order = min(len(ctrl_pts_1) - 1, MAX_SPLINE_ORDER)
    # Generate continuous spline from control points.
    tck, _ = interpolate.splprep(
        [ctrl_pts_1, ctrl_pts_2], s=SPLINE_SMOOTHNESS, k=spline_order
    )
    unew = np.linspace(0, 1, num=n)
    # Generate discrete spline
    # Numpy Array of dim (2, self.shape[1]) [x,r] / [x,y]
    shell_crest = np.asarray(interpolate.splev(unew, tck))

    return shell_crest



import numpy as np
from scipy.spatial.transform import Rotation as R

from kokiy.shell_2d import Shell2D, compute_normals_from_xyz
from kokiy.axishell import _compute_shell_crest


ATOL = 1e-8


class TorchShell(Shell2D):
    geom_type = "torch"

    def __init__(self, n_azi, n_longi, pt_a, pt_b, radius_a, radius_b, ecc_pt):
        super().__init__(n_azi, n_longi)

        self.pt_a = np.array(pt_a)
        self.pt_b = np.array(pt_b)
        self.ecc_pt = np.array(ecc_pt)
        self.radius_a = radius_a
        self.radius_b = radius_b

        # useful
        _vec_ab = self.pt_b - self.pt_a
        self.len_ab = np.linalg.norm(_vec_ab)
        self.axis_vect = _vec_ab / np.linalg.norm(_vec_ab)

        self._ecc = self.ecc_pt - self.pt_b
        self._verify_ecc()
        self._ecc_unit = self._ecc / np.linalg.norm(self._ecc)

        self._build_shell()

    def _verify_ecc(self):
        if abs(np.sum(self._ecc)) < ATOL and abs(np.prod(self._ecc)) < ATOL:
            raise Exception("Null eccentricities are not accepted")

        if abs(np.dot(self._ecc, self.axis_vect)) > ATOL:
            raise Exception("Eccentricity must be perpendicular to axis")

    def _build_shell(self):

        # Construct Shell Crest
        ctrl_pts_r = np.array([self._get_radius_sect(v) for v in self.v[0]])
        shell_crest = _compute_shell_crest(self.v[0], ctrl_pts_r, self.shape[1])

        # define directions
        center_vec = np.array([self.axis_vect * v for v in self.v[0]])

        # Compute radius and theta matrices of shape (n_azi, n_longi)
        _rad, _theta, _ = self._get_rad_theta_aux_normals(ctrl_pts_r)

        # get xyz
        xyz = []
        for i, sect_center in enumerate(center_vec):
            rotations = [
                R.from_rotvec(theta * self.axis_vect) for theta in _theta[:, i]
            ]
            rad_direcs = np.array(
                [rotation.apply(self._ecc_unit) for rotation in rotations]
            )

            xyz.append(
                sect_center
                + np.array(
                    [
                        sect_radius * rad_direc
                        for sect_radius, rad_direc in zip(_rad[:, i], rad_direcs)
                    ]
                )
            )

        self.x, self.y, self.z = [np.take(np.array(xyz), i, -1).T for i in range(3)]

        self.n_x, self.n_y, self.n_z = compute_normals_from_xyz(
            self.x, self.y, self.z, perio_u=True
        )

        shell_crest_scaled = shell_crest.copy()
        shell_crest_scaled[0] *= self.len_ab

        # Compute du, dv matrix of shape (n_azi, n_longi)
        longi_vals = np.tile(shell_crest_scaled[0], (self.shape[0], 1))
        self.du = np.pad(
            np.sqrt(np.diff(longi_vals, axis=1) ** 2 + np.diff(self.rad, axis=1) ** 2),
            ((0, 0), (1, 0)),
            "edge",
        )
        self.dv = _rad * np.pad(np.diff(_theta, axis=0), ((1, 0), (0, 0)), "edge")

        # Compute weight intervals in u and v directions array of shape (n_transvers, n_longi)
        self.dwu, self.dwv = self._compute_weight_interv_adim(self.du, self.dv)

    def _get_radius_sect(self, v):
        return (self.radius_b - self.radius_a) * v + self.radius_a

    def _get_eccentricity_sect(self, v):
        return self._ecc * v

    def _get_rad_theta_aux_normals(self, ctrl_pts_r):

        min_theta, max_theta = 0.0, np.deg2rad(360)
        theta_vec = np.linspace(min_theta, max_theta, num=self.shape[0])
        lim_theta_b = np.abs(np.arctan2(self.radius_b, np.linalg.norm(self._ecc)))

        # get phis (angles in circular region)
        phis = _get_phis(theta_vec, lim_theta_b)

        normal_90, normal_270 = self._get_planar_region_normals()

        rads = []
        thetas = []
        normals = []

        for sect_rad, v in zip(ctrl_pts_r, self.v[0]):
            sect_ecc = self._get_eccentricity_sect(v)
            sect_ecc_norm = np.linalg.norm(sect_ecc)
            lim_theta = np.abs(np.arctan2(sect_rad, sect_ecc_norm))
            corr_coeff = _compute_corr_coeff(lim_theta)

            rads_ = []
            thetas_ = []
            normals_ = []
            for phi, theta in zip(phis, theta_vec):
                if phi is None:  # rectangular region
                    theta_ = _get_theta_sect_rect(self.radius_b, v, theta, sect_rad)
                    rad_ = np.abs(sect_rad / np.sin(theta_))
                    normal_ = normal_90 if theta_ < np.pi else normal_270

                else:
                    theta_ = _get_theta_sect_via_phi(phi, corr_coeff)
                    rot_angle = phi[0]
                    rotation = R.from_rotvec(rot_angle * self.axis_vect)
                    radius_vec = self._ecc_unit * sect_rad

                    if phi[1] == np.pi:
                        vec = -sect_ecc + rotation.apply(-radius_vec)
                        normal_ = -rotation.apply(self._ecc_unit)
                    else:
                        vec = sect_ecc + rotation.apply(radius_vec)
                        normal_ = rotation.apply(self._ecc_unit)

                    rad_ = np.linalg.norm(vec)

                rads_.append(rad_)
                thetas_.append(theta_)
                normals_.append(normal_)

            rads.append(rads_)
            thetas.append(thetas_)
            normals.append(normals_)

        return (
            np.array(rads).T,
            np.array(thetas).T,
            np.array(normals).transpose(1, 0, 2),
        )

    def _get_planar_region_normals(self):
        rotations = [
            R.from_rotvec(rot_angle * self.axis_vect)
            for rot_angle in [np.pi / 2, 3 / 2 * np.pi]
        ]
        return [rotation.apply(self._ecc_unit) for rotation in rotations]

    def replicate(self, shape):
        return TorchShell(
            *shape, self.pt_a, self.pt_b, self.radius_a, self.radius_b, self.ecc_pt
        )


def _get_theta_sect_rect(radius_b, v, theta, sect_rad):
    """Gets horizontal distance in planar region."""
    horiz_dist = v * radius_b / np.tan(theta)

    return abs(np.arctan2(sect_rad, horiz_dist)) + _get_theta_add(theta)


def _get_phis(theta_vec, lim_theta_b):
    """Get angles of rotation of right face.

    Within the circular region, these angles are kept among layers. Outside
    the circular region it returns None.
    """
    corr_coeff_b = _compute_corr_coeff(lim_theta_b)

    phis = []
    for theta in theta_vec:
        if theta < lim_theta_b:
            phi = (theta * corr_coeff_b, 0.0)
        elif theta > 2 * np.pi - lim_theta_b:
            phi = ((theta - 2 * np.pi) * corr_coeff_b, 2 * np.pi)
        elif theta > (np.pi - lim_theta_b) and theta < np.pi + lim_theta_b:
            phi = ((theta - np.pi) * corr_coeff_b, np.pi)
        else:
            phi = None
        phis.append(phi)
    return phis


def _compute_corr_coeff(lim_theta):
    return (np.pi / 2) / lim_theta


def _get_theta_sect_via_phi(phi, corr_coeff):
    return phi[0] / corr_coeff + phi[1]


def _get_theta_add(theta):
    add = 0.0 if theta < np.pi else np.pi
    return add



import os
import pytest
from pyvista.core.utilities import from_meshio
from pyhip.commands.mesh_idcard import extract_hdf_meshinfo
import yamio


HIP_NAME = {"tetra": "Tetrahedra", "hexahedron": "Hexahedra"}

ATOL = 1e-8


NOT_SUPPORTED = {
    "xml": set(["quad", "hexahedron"]),
}


@pytest.mark.parametrize("elem_type", ["tetra", "hexahedron"])
def test_thickshell_hip_export(tmpdir, param_thickaxishell, elem_type):
    """
    Notes:
        If mesh is read by hip, it is assumed to be working properly.

        Note verification of negative volumes is already done in HipWriter.
    """
    expected_patch_labels = ["Bottom", "Top", "Left", "Right", "Back", "Front"]

    filename = "shell.mesh.h5"
    _, shell = param_thickaxishell

    mesh_filename = os.path.join(tmpdir, filename)
    shell.export_mesh(mesh_filename, elem_type)

    mesh_info, _, _, patch_labels = extract_hdf_meshinfo(mesh_filename)

    # check element type
    assert list(mesh_info["Number of cells"].keys())[0] == HIP_NAME[elem_type]
    assert mesh_info["Number of boundary nodes"] > 0
    for patch_label, expected_patch_label in zip(patch_labels, expected_patch_labels):
        assert patch_label == expected_patch_label


@pytest.mark.parametrize("fmt", ["xdmf", "xml"])
@pytest.mark.parametrize("elem_type", ["tetra", "hexahedron"])
def test_thickshell_export(tmpdir, param_thickaxishell, elem_type, fmt):
    """
    Notes:
        Patches are lost in these formats.

        Writes and reads mesh. Uses pyvista to check volumes.
    """
    if elem_type in NOT_SUPPORTED.get(fmt, []):
        return

    filename = f"shell.{fmt}"
    _, shell = param_thickaxishell
    init_mesh = shell.get_mesh(elem_type)

    # write in given fmt
    mesh_filename = os.path.join(tmpdir, filename)
    shell.export_mesh(mesh_filename, elem_type)

    # read mesh
    mesh = yamio.read(mesh_filename)

    # compare meshes
    # TODO: update when yamio has new functionaties
    mesh = yamio.Mesh(mesh.points, mesh.cells)
    init_mesh.bnd_patches = {}
    assert mesh == init_mesh

    # check areas (if positive)
    cell_volumes = from_meshio(mesh).compute_cell_sizes().cell_data["Volume"]
    assert all(cell_volumes > 0.0)


@pytest.mark.parametrize("fmt", ["xdmf", "xml"])
@pytest.mark.parametrize("elem_type", ["triangle", "quad"])
def test_shell2d_export(tmpdir, shell_2d, elem_type, fmt):
    """
    Notes:
        Writes and reads mesh, looking to the points and cells.
    """
    if elem_type in NOT_SUPPORTED.get(fmt, []):
        return

    shell_type, shell = shell_2d
    if shell_type in ["torchshell"]:  # torch has null areas
        return

    filename = f"shell.{fmt}"

    # write in given fmt
    mesh_filename = os.path.join(tmpdir, filename)
    init_mesh = shell.get_mesh(elem_type)
    yamio.write(mesh_filename, init_mesh)

    # read back
    mesh = yamio.read(mesh_filename)

    # compare meshes
    # TODO: update when yamio has new functionaties
    mesh = yamio.Mesh(mesh.points, mesh.cells)
    init_mesh.bnd_patches = {}
    assert mesh == init_mesh

    # check areas (if positive)
    cell_areas = from_meshio(mesh).compute_cell_sizes().cell_data["Area"]
    assert all(cell_areas > 0.0)



"""Defines a frustum shell object.
"""

import numpy as np
from scipy.spatial.transform import Rotation
from kokiy.axishell import AxiShell


class FrustumShell(AxiShell):
    """A frustum computational shell.

    Args:
        n_azi (int): Number of azimuthal shell points.
        n_longi (int): Number of longitudinal shell points.
        pt_a (np.array): coordinates of left circle's center.
        pt_b (np.array): coordinates of right circle's center.
        radius_a (float): left circle's radius.
        radius_b (float): right circle's radius.
        direc_pt (np.array): Deprecated
    """

    geom_type = "frustum"

    def __init__(self, n_azi, n_longi, pt_a, pt_b, radius_a, radius_b, direc_pt=None):

        # Create an AxiShell of same Shape, then rotate and translate.
        vect_b = np.array(pt_b)
        vect_a = np.array(pt_a)
        vect_a_b = vect_b - vect_a
        len_ab = np.linalg.norm(vect_a_b)
        vect_a_b /= len_ab

        super().__init__(n_azi, n_longi, 360.0, (0.0, len_ab), (radius_a, radius_b))

        rotation = _find_rotation(vect_a_b)
        self.x, self.y, self.z = _rotate_xyz(rotation, self.x, self.y, self.z)
        self.n_x, self.n_y, self.n_z = _rotate_xyz(
            rotation, self.n_x, self.n_y, self.n_z
        )

        self.x += vect_a[0]
        self.y += vect_a[1]
        self.z += vect_a[2]


def _find_rotation(vect_a_b):
    """Find the scipy rotation from vect X to vect AB"""
    rot_vect = np.cross((1.0, 0.0, 0.0), vect_a_b)
    parallelogram = np.linalg.norm(rot_vect)
    if parallelogram == 0.0:
        rot_vect *= 0
    else:
        angle = np.arcsin(parallelogram)
        rot_vect *= angle / parallelogram
    return Rotation.from_rotvec(rot_vect)


def _rotate_xyz(rotation, x, y, z):
    """Apply a scipty rotation on vects x,y,z"""
    shape_ = x.shape
    xyz = np.stack(
        [
            x.ravel(),
            y.ravel(),
            z.ravel(),
        ],
        axis=1,
    )

    new_xyz = rotation.apply(xyz)

    new_x = new_xyz[:, 0].reshape(shape_)
    new_y = new_xyz[:, 1].reshape(shape_)
    new_z = new_xyz[:, 2].reshape(shape_)

    return new_x, new_y, new_z



"""Defines a cartesian shell object.
"""

import numpy as np

from kokiy.shell_2d import Shell2D


class CloudShell(Shell2D):
    r"""Base class for Shell of Cloud points.

    This is a degeneration of the usual Shell2D.
    Each point in this point cloud is repeated in the u-direction (u=0 and u=1)
    to keep the shell2D. This avoid many ifs in the code.
    The U and V connectivity here is just random

    The CLoudShell, and its ThickCloudShell expansion
    should be presented as Points.
    """

    geom_type = "cloud"

    def __init__(self, xyz, normals, surf):

        size = xyz.shape[0]
        super().__init__(2, size)
        self.x = np.array([xyz[:, 0], xyz[:, 0]])
        self.y = np.array([xyz[:, 1], xyz[:, 1]])
        self.z = np.array([xyz[:, 2], xyz[:, 2]])
        self.n_x = np.array([normals[:, 0], normals[:, 0]])
        self.n_y = np.array([normals[:, 1], normals[:, 1]])
        self.n_z = np.array([normals[:, 2], normals[:, 2]])

        # Keept surface element coherent
        sqrsurf = np.sqrt(surf)
        self.du = np.array([sqrsurf, sqrsurf])
        self.dv = np.array([sqrsurf, sqrsurf])
        self.dwu = self.du
        self.dwv = self.dv

    # @property
    # def abs_curv(self):
    #     """override abs curvi"""
    #     return 1.

    def replicate(self, shape):
        """override replicate"""
        raise NotImplementedError("replicate not avialable for CloudShells")



"""Defines a 3D ThickShell object built by extruding 2D shells.
"""

import numpy as np
from scipy import interpolate

import meshio

from kokiy.shell import Shell
from kokiy._mesh_utils import get_bound_nodes
from kokiy._structured import get_uv
from kokiy._structured import interp_on_uv


class ThickShell(Shell):
    """A 3D shell object build by extruding a xyz mesh in the normals direction.

    Args:
        init_xyz (np.array): 2d mesh (e.g. from Shell2D).
        normals (np.array): 2d mesh normals with same shame as `init_xyz`.
        width_points (np.array): Width profile or width matrix.
        n_layers (int): Number of layers.
        shift (float): Additional depth.

    Attributes:
        shape (array-like): Shape of the shell.
    """

    geom_type = "thick"

    def __init__(self, init_xyz, normals, width_points, n_layers=2, shift=0.0):

        shape = init_xyz.shape[:-1] + (n_layers,)
        super().__init__(shape)

        self.init_xyz = init_xyz
        self.normals = normals
        self.width_points = np.array(width_points)
        self.shift = shift
        # compute variables
        self._xyz, self._dz = self._bake()  # private for the deprecated object
        # to know where bottom is
        self._neg_dir = False if np.sign(width_points[0][-1]) == 1.0 else True

    @property
    def xyz(self):
        """xyz matrix of shape (`*self.shape`, 3)."""
        return self._xyz

    @property
    def dz(self):
        """z-direction spacing matrix of shape `self.shape`."""
        return self._dz

    def _bake(self):
        n_layers = self.shape[-1]
        if self.width_points.shape == self.shape[:2]:
            width_matrix = self.width_points
        else:
            width_matrix = create_width_profile(self.width_points, self.shape[:2])

        xyz = self._compute_xyz(width_matrix, n_layers)
        dz = self._compute_dz(width_matrix, n_layers)

        return xyz, dz

    def _compute_xyz(self, width_matrix, n_layers):
        xyz = np.empty((*self.shape, 3))
        for j in range(3):
            n_j = np.take(self.normals, j, -1)
            for i in range(n_layers):
                xyz[:, :, i, j] = (
                    np.take(self.init_xyz, j, -1)
                    + (1.0 * i / n_layers * width_matrix + self.shift) * n_j
                )

        return xyz

    def _compute_dz(self, width_matrix, n_layers):
        dz = np.empty(self.shape)
        for i in range(n_layers):
            dz[:, :, i] = abs(width_matrix / n_layers)
        dz[:, :, 0] /= 2
        dz[:, :, -1] /= 2

        return dz

    @classmethod
    def bake_from_shell(cls, shell, width_points, n_layers=1, shift=0.0):
        return ThickShell(
            shell.xyz, shell.n_xyz, width_points, n_layers=n_layers, shift=shift
        )

    def dump(self, filename="shell", fields=None):
        """Dump shell geometrical and physical properties into hdf file.

        Args:
            filename (str): Name of the file to dump.
            fields (dict): Additional fields to dump. (key, `np.array`).
        """
        var_names = ["dz"]
        self._dump(filename, var_names, fields=fields)

    def dump_shell(self, name="shell", fields=None):
        """Dump shell geometrical properties into hdf file.

        Args:
            name (str): Name of the file to dump. Default 'shell'.
            fields (dict): Additional fields to dump. (key, `np.array`).

        .. deprecated:: 0.2.0
            Use `dump` instead.
        """
        self.dump(filename=name, fields=fields)

    def _get_bound_nodes(self):
        return get_bound_nodes(*self.shape, reverse=self._neg_dir)

    def _get_bound_nodes_display(self, **kwargs):
        return self._get_bound_nodes()

    def _get_points(self):
        """Returns raveled points."""
        points = np.empty((0, 3))
        for k in range(self.shape[2]):
            layer_coords = np.take(self.xyz, k, axis=2)
            coords = np.array(
                [np.take(layer_coords, i, axis=-1).ravel() for i in range(3)]
            ).T
            points = np.r_[points, coords]

        return points

    def _get_cells(self, elem_type, **kwargs):
        return super()._get_cells(elem_type, reverse=self._neg_dir)

    def _reshape_simplified_bnd_conns(self, conns, show_all):
        size = conns.size

        if size == (self.shape[0] * self.shape[1]):  # bottom or top
            new_conns = self._reshape_simplified_bnd_conns_helper(
                conns, self.shape[1], layered=False, show_all=show_all
            )
        else:
            if size == (self.shape[0] * self.shape[2]):  # laterals
                ref_shape = self.shape[0]
            else:  # back and front
                ref_shape = self.shape[1]
            new_conns = self._reshape_simplified_bnd_conns_helper(
                conns, ref_shape, layered=True, show_all=show_all
            )

        return [meshio.CellBlock("line", new_conns)]

    def _compute_replicate_info(self, shape):
        # interpolates mesh, normals and width_matrix (if applicable) to new shape
        shape_2d = shape[:2]

        init_shape = self.init_xyz.shape[:2]
        init_uv = get_uv(*init_shape, matrix_form=True).reshape(-1, 2)
        new_uv = get_uv(*shape_2d, matrix_form=True).reshape(-1, 2)

        init_xyz = self.init_xyz.reshape(-1, 3)
        new_xyz = interp_on_uv(init_uv, init_xyz, new_uv).reshape(*shape_2d, 3)

        init_normals = self.normals.reshape(-1, 3)
        new_normals = interp_on_uv(init_uv, init_normals, new_uv).reshape(*shape_2d, 3)

        if self.width_points.shape == init_shape:
            new_width_points = interp_on_uv(
                init_uv, self.width_points.reshape(-1), new_uv
            ).reshape(shape_2d)
        else:
            new_width_points = self.width_points

        return new_xyz, new_normals, new_width_points

    def replicate(self, shape):
        """Creates a similar instance, but possibly with different shape."""
        new_xyz, new_normals, new_width_points = self._compute_replicate_info(shape)

        return ThickShell(
            new_xyz, new_normals, new_width_points, n_layers=shape[-1], shift=self.shift
        )


class CakeDict(ThickShell, dict):
    """A ThickShell that keeps dict behavior.

    Use ThickShell instead.

    .. deprecated:: 0.2.0
        Use `ThickShell` instead.
    """

    def __init__(self, *args, **kwargs):
        ThickShell.__init__(self, *args, **kwargs)
        dict.__init__(self, xyz=self._xyz, dz=self._dz)
        self.fields = self

    @property
    def xyz(self):
        return self["xyz"]

    @property
    def dz(self):
        return self["dz"]

    @classmethod
    def bake_from_shell(cls, shell, width_points, n_layers=1, shift=0.0):
        return CakeDict(
            shell.xyz, shell.n_xyz, width_points, n_layers=n_layers, shift=shift
        )

    def dump(self, filename="shell", fields=None):
        """Dump shell geometrical and physical properties into hdf file.

        Args:
            filename (str): Name of the file to dump.
            fields (dict): Additional fields to dump. (key, `np.array`).
        """
        var_names = ["dz"]
        self._dump(
            filename, var_names, fields=fields, exclude_internal_fields=["xyz", "dz"]
        )

    def replicate(self, shape):
        """Creates a similar instance, but possibly with different shape."""
        new_xyz, new_normals, new_width_points = self._compute_replicate_info(shape)

        return CakeDict(
            new_xyz, new_normals, new_width_points, n_layers=shape[-1], shift=self.shift
        )


def create_width_profile(width_points, shape):

    (x_tuple, y_tuple) = tuple(zip(*width_points))

    # Extend Bounds
    xlist = [0] + list(x_tuple) + [1]
    ylist = [y_tuple[0]] + list(y_tuple) + [y_tuple[-1]]
    # Generate continuous fictive spline
    f_int = interpolate.interp1d(xlist, ylist)
    # Generate discretise spline
    xnew = np.linspace(0, 1, num=shape[1])
    ynew = f_int(xnew)

    return np.tile(ynew, (shape[0], 1))



import os

import pytest
import numpy as np

from kokiy.cartshell import CartShell
from kokiy.shell_utils import interpolate_solutions_on_shell
from kokiy.shell_utils import interpolate_solution_on_shell
from kokiy.shell_utils import infer_shell_from_xyz
from kokiy.shell_utils import shell_repr
from kokiy.shell_utils import shell_geom_type
from kokiy.shell_utils import _from_np_struct_to_dict


ATOL = 1e-6


def test_interpolate_solutions_on_shell_2d(datadir):

    # create fake shell to extract mesh
    n_trans, n_longi = 12, 24
    zero = np.array([0.00, 0.12, -0.2])
    u_max = np.array([0.00, 0.12, 0.2])
    v_max = np.array([0.20, 0.12, -0.2])
    init_shell = CartShell(n_trans, n_longi, zero, u_max, v_max)
    init_mesh = init_shell.xyz.reshape(-1, 3)

    # random data (a value for each node in the initial shell)
    np.random.seed(1)
    n_time, n_vars = 3, 2
    names = [f"random{i}" for i in range(n_vars)]
    dtype = dict(names=names, formats=[np.float32] * n_vars)
    shape = (n_time, init_mesh.shape[0])
    data = np.empty(shape, dtype=dtype)
    for name in names:
        data[name] = np.random.random(shape)

    # create new shell in which data will be interpolated
    n_trans, n_longi = 6, 12  # coarser than original
    zero = np.array([0.00, 0.12, -0.2])
    u_max = np.array([0.00, 0.12, 0.2])
    v_max = np.array([0.20, 0.12, -0.2])
    new_shell = CartShell(n_trans, n_longi, zero, u_max, v_max)

    # interpolate results (test array-like input)
    results = interpolate_solutions_on_shell(new_shell, init_mesh, data)

    # interpolate results (test dict-like input)
    data_dict = _from_np_struct_to_dict(data)
    results_dict = interpolate_solutions_on_shell(new_shell, init_mesh, data_dict)

    # compare with target
    filename = os.path.join(datadir, "interp_data.npz")
    target_results = np.load(filename, allow_pickle=True)
    for name in target_results.dtype.names:
        np.testing.assert_allclose(results[name], target_results[name], atol=ATOL)
        np.testing.assert_allclose(results_dict[name], target_results[name], atol=ATOL)

    # interpolate one solution interpolation
    var_name = "random0"
    result = interpolate_solution_on_shell(new_shell, init_mesh, data_dict[var_name])
    np.testing.assert_allclose(result, results_dict[var_name])

    # interpolate one solution at one time step only
    var_name = "random0"
    result = interpolate_solution_on_shell(new_shell, init_mesh, data_dict[var_name][0])
    np.testing.assert_allclose(result, results_dict[var_name][0])


def test_interpolate_solutions_on_shell_3d(thickshell):
    # just tests if it is able to interpolate itself
    # test on 2d ensures interpolation is not changing

    shell_type, init_shell = thickshell

    for name in ["frustum", "torch"]:  # TODO: make it work
        if name in shell_type:
            return

    new_shell = init_shell.replicate(init_shell.shape)

    init_mesh = init_shell.xyz.reshape(-1, 3)
    np.random.seed(1)
    data = np.random.random(
        (
            2,
            init_mesh.shape[0],
        )
    )

    # interpolate
    new_data = interpolate_solution_on_shell(new_shell, init_mesh, data)

    np.testing.assert_allclose(new_data.reshape(2, -1), data, atol=ATOL)


def test_cartshell_create_structured_base():
    """
    Test of generation of cartshell
    """

    n_y = 3
    n_z = 3

    mesh = _create_mesh()

    shell = infer_shell_from_xyz(mesh, "cart", (n_y, n_z))
    xyz_tgt = np.array(
        [
            [
                [0.0, 0.0, 0.0],
                [0.0, 0.5, 0.0],
                [0.0, 1.0, 0.0],
            ],
            [
                [0.0, 0.0, 0.5],
                [0.0, 0.5, 0.5],
                [0.0, 1.0, 0.5],
            ],
            [
                [0.0, 0.0, 1.0],
                [0.0, 0.5, 1.0],
                [0.0, 1.0, 1.0],
            ],
        ]
    )
    np.testing.assert_allclose(xyz_tgt, shell.xyz)


def test_axishell_create_structured_based():
    # TODO: need to test bnd_uv

    mesh = _create_mesh()

    n_r = 3
    n_theta = 3
    shell = infer_shell_from_xyz(mesh, "axicyl", (n_r, n_theta))
    xyz_tgt = np.array(
        [
            [
                [0.0, 7.08085476e-01, 0.0],
                [0.0, 8.54042738e-01, 0.0],
                [0.0, 1.0, 0.0],
            ],
            [
                [0.0, 5.00692042e-01, 5.00692042e-01],
                [0.0, 6.03899412e-01, 6.03899412e-01],
                [0.0, 7.07106781e-01, 7.07106781e-01],
            ],
            [
                [0.0, 0.0, 7.08085476e-01],
                [0.0, 0.0, 8.54042738e-01],
                [0.0, 0.0, 1.0],
            ],
        ]
    )
    np.testing.assert_allclose(xyz_tgt, shell.xyz, atol=ATOL)


@pytest.mark.parametrize(
    "load_shell_case",
    [
        ("cartshell", "cart"),
        ("axishell", "axicyl"),
        ("torch_shell", "torch"),
        ("frustum_shell", "frustum"),
    ],
    indirect=True,
)
def test_shell_repr(load_shell_case):
    # TODO: will be deprecated -> move test to within each of the objects
    (_, shell), geom_type = load_shell_case

    target_value = f"Shell\n-----\n\nshape : (8, 10)\ntype : {geom_type}"
    assert shell_repr(shell) == target_value
    assert repr(shell) == target_value


@pytest.mark.parametrize(
    "load_shell_case",
    [
        ("cartshell", "cart"),
        ("axishell", "axicyl"),
        ("torch_shell", "torch"),
        ("frustum_shell", "frustum"),
    ],
    indirect=True,
)
def test_shell_geom_type(load_shell_case):
    # TODO: will be deprecate (delete)
    (_, shell), geom_type = load_shell_case
    assert shell_geom_type(shell) == geom_type


def _create_mesh():
    yi = np.linspace(0, 1, 20)
    zi = np.linspace(1, 0, 20)
    xi = np.zeros_like(yi)

    return np.stack((xi, yi, zi), axis=-1)
